# -------------------------------------------------------------------- #
#   EntityRef.pm - Dereference Entity
#   Copyright (C) 2000-2004 Kawasaki Yuusuke <u-suke@kawa.net>
# -------------------------------------------------------------------- #
#   2004/10/22  
#   2004/11/07  POD 򾯤ޤ
# -------------------------------------------------------------------- #
    package EntityRef;
    use strict;
    use vars qw( $VERSION $DEBUG );
    $VERSION = "0.01";
# -------------------------------------------------------------------- #
=head1 NAME

EntityRef.pm - Dereference Entity

=head1 SYNOPSIS

    use EntityRef;
    my $text = "EURO &euro; ALPHA &Alpha; COPY &copy;\n";
    my $er = new EntityRef();
    $er->entity2hex( \$text );
    print $text, "\n";

=head1 DESCRIPTION

=head2 new()

    HTML/XML ѤμλȤޥåץեɤ߹ߤޤ
    ǥեȤǤ HTMLlat1.entHTMLspecial.entHTMLsymbol.ent 
    Ĥɸޥåץեɤ߹ߤޤ
    Υե EntityRef.pm Ʊǥ쥯ȥ֤Ƥ
    new() 裱ޥåץեꤹ뤳ȤǽǤ

        my $er = new EntityRef( "EmojiDoCoMo.ent" );

    ɥ⳨ʸѤμλȤޥåץեΤߤɤ߹ߤޤ

=head2 entity2hex()

    λɽUnicodeֹɽ(16ʿ)Ѵޤ

    HTMLspecial &euro;  ->  &#x20AC;    -- euro sign, U+20AC NEW
    HTMLsymbol  &Alpha; ->  &#x0391;    -- greek capital letter alpha, U+0391
    HTMLlat1    &copy;  ->  &#x00A9;    -- copyright sign, U+00A9 ISOnum

=head2 .ent

    λȤޥåץեϡηȤʤޤ

    <!ENTITY ̾ CDATA "&#;" --  -->

    ȤʣԤϤäƤ⹽ޤ
    ̾ʸʸ̤Ƥޤ
    ʸɤ10ʿޤ16ʿ Unicode ǻꤷƤ
    ʣʸϤλȤŸǤޤ󡣡1ʸΤŸޤ
    ޤʸֹ椬 128ʲʸŸޤ

=head1 SEE ALSO

Extensible Markup Language (XML) 1.0
http://www.w3.org/TR/REC-xml

    4.1 Character and Entity References

    [66]  CharRef  ::=  '&#' [0-9]+ ';'  
                        | '&#x' [0-9a-fA-F]+ ';'
    [67]  Reference  ::=  EntityRef | CharRef 
    [68]  EntityRef  ::=  '&' Name ';'

    If the character reference begins with "&#x", the digits and letters 
    up to the terminating ; provide a hexadecimal representation of the 
    character's code point in ISO/IEC 10646. If it begins just with "&#", 
    the digits up to the terminating ; provide a decimal representation 
    of the character's code point.

http://www.w3.org/TR/html401/HTMLlat1.ent       -- Latin-1 entities 
http://www.w3.org/TR/html401/HTMLsymbol.ent     -- Symbol entities 
http://www.w3.org/TR/html401/HTMLspecial.ent    -- Special entities 

=head1 COPYRIGHT

    Copyright 2004 Kawasaki Yusuke <u-suke@kawa.net>
    http://www.kawa.net/

=cut
# -------------------------------------------------------------------- #
#   ǥեȤɤ߽Ф .ent
# -------------------------------------------------------------------- #
    my $DEFAULT_ENTITIES = [qw(
        HTMLlat1.ent
        HTMLspecial.ent
        HTMLsymbol.ent
    )];
# -------------------------------------------------------------------- #
#   󥹥ȥ饯
# -------------------------------------------------------------------- #
sub new {
    my $package = shift or return;
    my $names = [ @_ ];

    my $obj = {};
    bless $obj, $package;

    my $pwd = &where_you_are();                 # ǥ쥯ȥ
    unless ( scalar @$names ) {
        # $names = &entity_list( $pwd );        # ƤФ
        $names = $DEFAULT_ENTITIES;
        unless ( ref $names && scalar @$names ) {
            die "no .ent files found at '$pwd'\n";
        }
    }

    foreach my $name1 ( @$names ) {
        my $ent;
        if ( $name1 =~ m#^/# ) {
            $obj->read_entity( $name1 );    # Хѥ
        } else {
            my $file = $pwd ."/". $name1;
            $obj->read_entity( $file ); # Хѥ
        }
    }

    $obj;                                   # ֥Ȥ֤
}
# -------------------------------------------------------------------- #
#   &#xHHHH; ͤĥϥå֤᥽å
# -------------------------------------------------------------------- #
sub get_hex_hash {
    my $obj = shift;
    my $hash1 = $obj->{entity};
    my $hash2 = { map {$_=>sprintf("&#x%04X;",$hash1->{$_})} keys %$hash1 };
    $hash2;
}
# -------------------------------------------------------------------- #
#   ǥ쥯ȥƤ .ent ե̾ФʸߤԻѡ
# -------------------------------------------------------------------- #
sub entity_list {
    my $path = shift;

    opendir( DIR, $path ) or die "$! '$path'\n";
    my $files = [];
    while( 1 ) {
        my $file = readdir( DIR );
        last unless defined $file;
        next unless ( $file =~ m#[^\/]\.ent$#i );
        push( @$files, $file );
    }
    closedir( DIR );
    $files;
}
# -------------------------------------------------------------------- #
#   .ent եɤ߹0x00800xFFFDޤǡ
# -------------------------------------------------------------------- #
sub read_entity {
    my $obj = shift;
    my $file = shift;
    my $hash = $obj->{entity} || {};
#   $DEBUG and print "[read_entity] $file\n";
    open( ENT, $file ) or die "$! '$file'\n";
    local $/ = undef;
    my $body = join( "", <ENT> );
    while( $body =~ m/<!ENTITY\s+(\w+)\s+CDATA\s+
                                 (?:([^\s"]+)|"([^"]+)")[^>]+>/xg ) {
        my( $ref, $entity ) = ( $1, $2||$3 );
        my( $code, $hex ) = ( $entity =~ m/^\&\#(?:(\d+)|x([0-9A-Fa-f]+));$/s );
        $code = hex($hex) if ( $hex ne "" );
        next if ( $code < 0x0080 );
        next if ( $code >= 0xFFFE );
        # $DEBUG and print "[read_entity] $file &$ref; $code\n";
        $hash->{$ref} = $code;
    }
    close( ENT );
    $obj->{entity} = $hash;
}
# -------------------------------------------------------------------- #
#   λɽ򥳡ɻɽѴ
# -------------------------------------------------------------------- #
sub entity2hex {
    my $obj = shift;
    my $src = \$_[0];
    $src = $$src if ref $$src;              # ɬ٥Υե
    my $ent = $obj->{entity} or return;
    $$src =~ s{ (\&([A-Za-z][\w\-]*);) }
              { defined $ent->{$2} ? sprintf("&#x%04X;",$ent->{$2}) : $1 }gex;
}
# -------------------------------------------------------------------- #
#   ¹Υե̾Ф
# -------------------------------------------------------------------- #
sub where_you_are {
    my $path = (caller(0))[1];      # ƤӽФ¹Υե̾
    $path =~ s#[^/]+$##s;           # ǥ쥯ȥ̾Ѵ
    $path = "." if ( $path eq "" );
    $path;                          # ǥ쥯ȥ֤̾
}
# -------------------------------------------------------------------- #
;1; # End of the script.
# -------------------------------------------------------------------- #
